class NpotPatch {
  static applyEncryptedExt(path) {
    if (!path.endsWith("png")) return path;
    if (typeof Decrypter === "undefined") {
      console.warn("applyEncryptedExt: Decrypter not found, appending '_'");
      return path + "_";
    }
    if (Decrypter.hasEncryptedImages) {
      if (Decrypter.checkImgIgnore(path)) {
        return path;
      }
      return Decrypter.extToEncryptExt(path);
    }
    return path;
  }

  static npotInfo(imageSource) {
    if (Object.keys(NpotPatch.npotInfoCache).length > 256) {
      NpotPatch.npotInfoCache = {};
    }

    if (NpotPatch.npotInfoCache[imageSource]) {
      return NpotPatch.npotInfoCache[imageSource];
    }

    if (window._ImageInterface) {
      NpotPatch.dummyAnchor.href = imageSource;
      let href = NpotPatch.dummyAnchor.href;

      try {
        const rawNameResult = JSON.parse(
          _ImageInterface.getOriginalSize(new URL(href).pathname)
        );
        if (rawNameResult instanceof Array && rawNameResult.length === 2) {
          return (NpotPatch.npotInfoCache[imageSource] = rawNameResult);
        }
      } catch (err) {
        console.error("npotInfo error (raw):", err);
      }
      try {
        href = NpotPatch.applyEncryptedExt(href);

        return (NpotPatch.npotInfoCache[imageSource] = JSON.parse(
          _ImageInterface.getOriginalSize(new URL(href).pathname)
        ));
      } catch (err) {
        console.error("npotInfo error (encrypted):", err);
      }
    } else {
      console.warn("npotInfo: _ImageInterface not found, using fallback");
      try {
        const img = new Image();
        img.src = imageSource;
        const dimension = [img.width, img.height];
        if (dimension[0] > 0 && dimension[1] > 0) {
          console.log("npotInfo: Fallback dimensions", dimension);
          return (NpotPatch.npotInfoCache[imageSource] = dimension);
        } else {
          console.warn("npotInfo: Invalid fallback dimensions", dimension);
        }
      } catch (err) {
        console.error("npotInfo fallback error:", err);
      }
    }
    return undefined;
  }

  static immediatelyGetInfo(imageSource) {
    try {
      NpotPatch.dummyAnchor.href = imageSource;
      let href = NpotPatch.dummyAnchor.href;

      href = NpotPatch.applyEncryptedExt(href);

      if (NpotPatch.npotInfoCache[href]) {
        return NpotPatch.npotInfoCache[href];
      }

      const xhr = new XMLHttpRequest();
      xhr.open("GET", href, false);
      xhr.send();

      return NpotPatch.npotInfo(href);
    } catch(err) {
      console.error("immediatelyGetInfo error:", err);
      return undefined;
    }
  }

  static getShrinkSize([width = 1, height = 1]) {
    const resizedWidth = Math.min(width, 2048);
    const resizedHeight = Math.min(height, 2048);
    return [resizedWidth, resizedHeight];
  }

  static getShrinkRatio([width = 1, height = 1]) {
    const [resizedWidth, resizedHeight] = NpotPatch.getShrinkSize([
      width,
      height,
    ]);

    return [resizedWidth / width, resizedHeight / height];
  }

  static patchBitmap() {
    if (!window.Bitmap) {
      console.error("patchBitmap: Bitmap not found");
      return;
    }
    const _oldLoad = Bitmap.prototype._onLoad;
    Bitmap.prototype._onLoad = function () {
      _oldLoad.call(this);

      if (this._image == null) {
        console.warn("patchBitmap: _image is null");
        return;
      }
      try {
        const originalDimension = NpotPatch.npotInfo(this._image.src);
        if (originalDimension && originalDimension.length === 2) {
          [this.originalWidth, this.originalHeight] = originalDimension;
          console.log("patchBitmap: Set original dimensions", originalDimension);
        }
      } catch (err) {
        console.error("patchBitmap error:", err);
      }
    };
  }

  static patchSprite() {
    if (!window.Sprite || !window.TilingSprite) {
      console.error("patchSprite: Sprite or TilingSprite not found");
      return;
    }
    try {
      const originalBitmapLoad = Sprite.prototype._onBitmapLoad;

      Sprite.prototype._onBitmapLoad = function (bitmapLoaded) {
        originalBitmapLoad.call(this, bitmapLoaded);
        if (!(bitmapLoaded instanceof Bitmap) || !bitmapLoaded._image) {
          console.warn("patchSprite: Invalid bitmap or _image");
          return;
        }
        if (bitmapLoaded === this._bitmap) {
          if (
            bitmapLoaded._image.width >= 2048 ||
            bitmapLoaded._image.height >= 2048
          ) {
            const dimension = NpotPatch.npotInfo(bitmapLoaded._url);
            if (!dimension || !(dimension instanceof Array) || dimension.length !== 2) {
              console.warn("patchSprite: Invalid dimension", dimension);
              return;
            }
            const [widthRatio, heightRatio] = NpotPatch.getShrinkRatio(dimension);
            if (widthRatio <= 0 || heightRatio <= 0) {
              console.warn("patchSprite: Invalid ratios", [widthRatio, heightRatio]);
              return;
            }
            [this.width, this.height] = dimension;
            this.scale = new PIXI.Point(1 / widthRatio, 1 / heightRatio);
            console.log("patchSprite: Scaled sprite", dimension, [widthRatio, heightRatio]);
          }
        }
      };

      const originalTilingBitmapLoad = TilingSprite.prototype._onBitmapLoad;

      TilingSprite.prototype._onBitmapLoad = function (bitmapLoaded) {
        originalTilingBitmapLoad.call(this, bitmapLoaded);
        if (!(bitmapLoaded instanceof Bitmap) || !bitmapLoaded._image) {
          console.warn("patchTilingSprite: Invalid bitmap or _image");
          return;
        }
        if (
          bitmapLoaded._image.width >= 2048 ||
          bitmapLoaded._image.height >= 2048
        ) {
          const dimension = NpotPatch.npotInfo(bitmapLoaded._url);
          if (!dimension || !(dimension instanceof Array) || dimension.length !== 2) {
            console.warn("patchTilingSprite: Invalid dimension", dimension);
            return;
          }
          const [widthRatio, heightRatio] = NpotPatch.getShrinkRatio(dimension);
          if (widthRatio <= 0 || heightRatio <= 0) {
            console.warn("patchTilingSprite: Invalid ratios", [widthRatio, heightRatio]);
            return;
          }
          this.tileScale = new PIXI.Point(1 / widthRatio, 1 / heightRatio);
          console.log("patchTilingSprite: Scaled tile", dimension, [widthRatio, heightRatio]);
        }
      };
    } catch (err) {
      console.error("patchSprite error:", err);
    }
  }

  static patchPixiNpot() {
    if (!window.PIXI) {
      console.error("patchPixiNpot: PIXI not found");
      return;
    }
    function handlePixiTexture(baseTexture, frame) {
      if (!baseTexture) {
        console.warn("handlePixiTexture: baseTexture is null");
        return;
      }

      let originalDimension;
      if (baseTexture.originalDimension) {
        originalDimension = baseTexture.originalDimension;
      } else if (baseTexture.width >= 2048 || baseTexture.height >= 2048) {
        const baseTextureSource = (() => {
          if (baseTexture.resource) {
            if (baseTexture.resource.source) {
              return baseTexture.resource.source;
            }
          }
          return baseTexture.source;
        })();

        if (!baseTextureSource) {
          console.warn("handlePixiTexture: baseTextureSource is null");
          return;
        }

        if (baseTextureSource instanceof HTMLCanvasElement) {
          const newCanvas = document.createElement("canvas");
          newCanvas.width = Math.min(baseTextureSource.width, 2048);
          newCanvas.height = Math.min(baseTextureSource.height, 2048);
          const ctx = newCanvas.getContext("2d");
          ctx.drawImage(
            baseTextureSource,
            0,
            0,
            newCanvas.width,
            newCanvas.height
          );

          baseTexture.originalDimension = originalDimension = [
            baseTextureSource.width,
            baseTextureSource.height,
          ];

          if (baseTexture.resource) {
            if (baseTexture.resource.source) {
              baseTexture.resource.source.width = newCanvas.width;
              baseTexture.resource.source.height = newCanvas.height;
            } else {
              baseTexture.source.width = newCanvas.width;
              baseTexture.source.height = newCanvas.height;
            }
          } else {
            baseTexture.source.width = newCanvas.width;
            baseTexture.source.height = newCanvas.height;
          }

          const bctx = baseTextureSource.getContext("2d");
          bctx.drawImage(
            newCanvas,
            0,
            0,
            baseTextureSource.width,
            baseTextureSource.height
          );

          baseTexture.width = newCanvas.width;
          baseTexture.height = newCanvas.height;
          console.log("handlePixiTexture: Resized canvas", [baseTexture.width, baseTexture.height]);
        } else {
          originalDimension = NpotPatch.npotInfo(baseTextureSource.src);
        }
      }

      if (
        originalDimension instanceof Array &&
        originalDimension.length === 2
      ) {
        if (
          frame.width === baseTexture.width &&
          frame.height === baseTexture.height
        )
          return;
        const [originalWidth, originalHeight] = originalDimension;

        frame.x = (frame.x / originalWidth) * baseTexture.width;
        frame.y = (frame.y / originalHeight) * baseTexture.height;
        frame.width = (frame.width / originalWidth) * baseTexture.width;
        frame.height = (frame.height / originalHeight) * baseTexture.height;
        console.log("handlePixiTexture: Adjusted frame", frame);
      }
    }

    try {
      const pixiTexture = window.Texture || PIXI.Texture;
      if (!pixiTexture) {
        console.error("patchPixiNpot: pixiTexture not found");
        return;
      }

      const originalSetter = Object.getOwnPropertyDescriptor(
        pixiTexture.prototype,
        "frame"
      ).set;

      Object.defineProperty(pixiTexture.prototype, "frame", {
        set: function (frame) {
          handlePixiTexture(this.baseTexture, frame);
          originalSetter.call(this, frame);
        },
      });
      console.log("patchPixiNpot: Successfully patched PIXI texture frame");
    } catch (err) {
      console.error("patchPixiNpot error:", err);
    }
  }

  static resizeSystemIcon([widthRatio = 1, heightRatio = 1]) {
    if (!window.ImageManager || !window.Window_Base) {
      console.error("resizeSystemIcon: ImageManager or Window_Base not found");
      return;
    }
    ImageManager._iconWidth *= widthRatio;
    ImageManager._iconHeight *= widthRatio;

    Window_Base._iconWidth *= widthRatio;
    Window_Base._iconHeight *= widthRatio;

    Window_Base._iconSrcWidth *= widthRatio;
    Window_Base._iconSrcHeight *= widthRatio;
    console.log("resizeSystemIcon: Resized icons", [widthRatio, heightRatio]);
  }

  static patchSystemSpriteSets() {
    if (!window.Scene_Boot || !window.ImageManager) {
      console.error("patchSystemSpriteSets: Scene_Boot or ImageManager not found");
      return;
    }
    let spriteSetPatched = false;
    const oldReady = Scene_Boot.prototype.isReady;
    Scene_Boot.prototype.isReady = function () {
      const isReady = oldReady.call(this);
      if (isReady && !spriteSetPatched) {
        spriteSetPatched = true;

        const iconSet = ImageManager.loadSystem("IconSet");
        const iconSetUrl = iconSet.url;

        const originalDimension = NpotPatch.immediatelyGetInfo(iconSet._image.src);

        if (
          originalDimension instanceof Array &&
          originalDimension.length === 2
        ) {
          const shrinkRatio = NpotPatch.getShrinkRatio(originalDimension);
          NpotPatch.resizeSystemIcon(shrinkRatio);

          const oldBlt = Bitmap.prototype.blt;

          Bitmap.prototype.blt = function (
            source,
            sx,
            sy,
            sw,
            sh,
            dx,
            dy,
            dw,
            dh
          ) {
            if (source._url === iconSetUrl) {
              dw = dw || sw / shrinkRatio[0];
              dh = dh || sh / shrinkRatio[1];
            }
            oldBlt.call(this, source, sx, sy, sw, sh, dx, dy, dw, dh);
          };
          console.log("patchSystemSpriteSets: Patched icon set", originalDimension, shrinkRatio);
        }
      }
      return isReady;
    };
  }

  static applyPatch(level) {
    if (this.isPatched) {
      console.warn("applyPatch: Already patched, skipping");
      return;
    }
    console.log(`Applying NPOT Patch with level: ${level}`);

    if (level === "none") {
      this.isPatched = true;
      return;
    }

    this.patchBitmap();
    this.patchSprite();

    if (level === "high") {
      this.patchPixiNpot();
      this.patchSystemSpriteSets();
    }

    this.isPatched = true;
  }

  static setTrigger(level) {
    console.log("setTrigger: Setting patch trigger with level", level);
    if (typeof Main !== "undefined") {
      const oldOnScriptLoad = Main.prototype.onScriptLoad;
      Main.prototype.onScriptLoad = function () {
        if (this.loadCount + 1 === this.numScripts) {
          console.log("setTrigger: Applying patch on script load");
          NpotPatch.applyPatch(level);
        }
        oldOnScriptLoad.call(this);
      };
    } else {
      console.log("setTrigger: Main undefined, applying patch immediately");
      NpotPatch.applyPatch(level);
    }
  }
}

NpotPatch.isPatched = false;
NpotPatch.npotInfoCache = {};
NpotPatch.checkedParallaxRatio = {};
NpotPatch.dummyAnchor = document.createElement("a");

// Function to be called from Android
window.applyNpotPatch = function(level) {
  console.log("applyNpotPatch: Called with level", level);
  NpotPatch.setTrigger(level);
};