class NpotPatch {
  static applyEncryptedExt(path) {
    if (!path.endsWith("png")) return path;
    if (typeof Decrypter === "undefined") {
      return path + "_";
    }
    if (Decrypter.hasEncryptedImages) {
      if (Decrypter.checkImgIgnore(path)) {
        return path;
      }
      return Decrypter.extToEncryptExt(path);
    }
    return path;
  }

  static npotInfo(imageSource) {
    if (Object.keys(NpotPatch.npotInfoCache).length > 256) {
      NpotPatch.npotInfoCache = {};
    }

    if (NpotPatch.npotInfoCache[imageSource]) {
      return NpotPatch.npotInfoCache[imageSource];
    } else {
      NpotPatch.dummyAnchor.href = imageSource;
      let href = NpotPatch.dummyAnchor.href;

      try {
        const rawNameResult = JSON.parse(
          _ImageInterface.getOriginalSize(new URL(href).pathname)
        );
        if (rawNameResult instanceof Array && rawNameResult.length === 2) {
          return (NpotPatch.npotInfoCache[imageSource] = rawNameResult);
        }
      } catch (err) {
        console.error(err);
      }
      try {
        href = NpotPatch.applyEncryptedExt(href);

        return (NpotPatch.npotInfoCache[imageSource] = JSON.parse(
          _ImageInterface.getOriginalSize(new URL(href).pathname)
        ));
      } catch (err) {}
    }
  }

  static immediatelyGetInfo(imageSource) {
    try {
      NpotPatch.dummyAnchor.href = imageSource;
      let href = NpotPatch.dummyAnchor.href;

      href = NpotPatch.applyEncryptedExt(href);

      if (NpotPatch.npotInfoCache[href]) {
        return NpotPatch.npotInfoCache[href];
      }

      const xhr = new XMLHttpRequest();
      xhr.open("GET", href, false);
      xhr.send();

      return NpotPatch.npotInfo(href);
    } catch(err) {
      console.error(err);
      return undefined;
    }
  }

  static getShrinkSize([width = 1, height = 1]) {
    const resizedWidth = Math.min(width, 4096);
    const resizedHeight = Math.min(height, 4096);

    return [resizedWidth, resizedHeight];
  }

  static getShrinkRatio([width = 1, height = 1]) {
    const [resizedWidth, resizedHeight] = NpotPatch.getShrinkSize([
      width,
      height,
    ]);

    return [resizedWidth / width, resizedHeight / height];
  }

  static patchBitmap() {
    const _oldLoad = Bitmap.prototype._onLoad;
    Bitmap.prototype._onLoad = function () {
      _oldLoad.call(this);

      if (this._image == null) {
        return;
      }
      const originalDimension = JSON.parse(
        _ImageInterface.getOriginalSize(new URL(this._image.src).pathname)
      );
      if (originalDimension.length === 2) {
        [this.originalWidth, this.originalHeight] = originalDimension;
      }
    };
  }

  static patchSprite() {
    try {
      const originalBitmapLoad = Sprite.prototype._onBitmapLoad;

      Sprite.prototype._onBitmapLoad = function (bitmapLoaded) {
        originalBitmapLoad.call(this, bitmapLoaded);
        if (!(bitmapLoaded instanceof Bitmap) || !bitmapLoaded._image) return;
        if (bitmapLoaded === this._bitmap) {
          if (
            bitmapLoaded._image.width >= 4096 ||
            bitmapLoaded._image.height >= 4096
          ) {
            const dimension = NpotPatch.npotInfo(bitmapLoaded._url);
            if (dimension instanceof Array && dimension.length === 2) {
              const [widthRatio, heightRatio] =
                NpotPatch.getShrinkRatio(dimension);
              [this.width, this.height] = dimension;
              this.scale = new PIXI.Point(1 / widthRatio, 1 / heightRatio);
            }
          }
        }
      };

      const originalTilingBitmapLoad = TilingSprite.prototype._onBitmapLoad;

      TilingSprite.prototype._onBitmapLoad = function (bitmapLoaded) {
        originalTilingBitmapLoad.call(this, bitmapLoaded);
        if (!(bitmapLoaded instanceof Bitmap) || !bitmapLoaded._image) return;
        if (
          bitmapLoaded._image.width >= 4096 ||
          bitmapLoaded._image.height >= 4096
        ) {
          const dimension = NpotPatch.npotInfo(bitmapLoaded._url);
          if (dimension instanceof Array && dimension.length === 2) {
            const [widthRatio, heightRatio] =
              NpotPatch.getShrinkRatio(dimension);
            this.tileScale = new PIXI.Point(1 / widthRatio, 1 / heightRatio);
          }
        }
      };
    } catch (err) {
      console.error(err);
    }
  }

  static patchPixiNpot() {
    function handlePixiTexture(baseTexture, frame) {
      if (!baseTexture) return;

      let originalDimension;
      if (baseTexture.originalDimension) {
        originalDimension = baseTexture.originalDimension;
      } else if (baseTexture.width >= 4096 || baseTexture.height >= 4096) {
        const baseTextureSource = (() => {
          if (baseTexture.resource) {
            if (baseTexture.resource.source) {
              return baseTexture.resource.source;
            }
          }
          return baseTexture.source;
        })();

        if (!baseTextureSource) return;

        if (baseTextureSource instanceof HTMLCanvasElement) {
          const newCanvas = document.createElement("canvas");
          newCanvas.width = Math.min(baseTextureSource.width, 4096);
          newCanvas.height = Math.min(baseTextureSource.height, 4096);
          const ctx = newCanvas.getContext("2d");
          ctx.drawImage(
            baseTextureSource,
            0,
            0,
            newCanvas.width,
            newCanvas.height
          );

          baseTexture.originalDimension = originalDimension = [
            baseTextureSource.width,
            baseTextureSource.height,
          ];

          if (baseTexture.resource) {
            if (baseTexture.resource.source) {
              baseTexture.resource.source.width = newCanvas.width;
              baseTexture.resource.source.height = newCanvas.height;
            } else {
              baseTexture.source.width = newCanvas.width;
              baseTexture.source.height = newCanvas.height;
            }
          } else {
            baseTexture.source.width = newCanvas.width;
            baseTexture.source.height = newCanvas.height;
          }

          const bctx = baseTextureSource.getContext("2d");
          bctx.drawImage(
            newCanvas,
            0,
            0,
            baseTextureSource.width,
            baseTextureSource.height
          );

          baseTexture.width = newCanvas.width;
          baseTexture.height = newCanvas.height;
        } else {
          originalDimension = NpotPatch.npotInfo(baseTextureSource.src);
        }
      }

      if (
        originalDimension instanceof Array &&
        originalDimension.length === 2
      ) {
        if (
          frame.width === baseTexture.width &&
          frame.height === baseTexture.height
        )
          return;
        const [originalWidth, originalHeight] = originalDimension;

        frame.x = (frame.x / originalWidth) * baseTexture.width;
        frame.y = (frame.y / originalHeight) * baseTexture.height;
        frame.width = (frame.width / originalWidth) * baseTexture.width;
        frame.height = (frame.height / originalHeight) * baseTexture.height;
      }
    }

    try {
      const pixiTexture = window.Texture || PIXI.Texture;

      const originalSetter = Object.getOwnPropertyDescriptor(
        pixiTexture.prototype,
        "frame"
      ).set;

      Object.defineProperty(pixiTexture.prototype, "frame", {
        set: function (frame) {
          handlePixiTexture(this.baseTexture, frame);
          originalSetter.call(this, frame);
        },
      });
    } catch (err) {
      console.error(err);
    }
  }

  static resizeSystemIcon([widthRatio = 1, heightRatio = 1]) {
    ImageManager._iconWidth *= widthRatio;
    ImageManager._iconHeight *= heightRatio;

    Window_Base._iconWidth *= widthRatio;
    Window_Base._iconHeight *= heightRatio;

    Window_Base._iconSrcWidth *= widthRatio;
    Window_Base._iconSrcHeight *= heightRatio;
  }

  static patchSystemSpriteSets() {
    let spriteSetPatched = false;
    const oldReady = Scene_Boot.prototype.isReady;
    Scene_Boot.prototype.isReady = function () {
      const isReady = oldReady.call(this);
      if (isReady && !spriteSetPatched) {
        spriteSetPatched = true;

        const iconSet = ImageManager.loadSystem("IconSet");
        const iconSetUrl = iconSet.url;

        const originalDimension = NpotPatch.immediatelyGetInfo(iconSet._image.src);

        if (
          originalDimension instanceof Array &&
          originalDimension.length === 2
        ) {
          const shrinkRatio = NpotPatch.getShrinkRatio(originalDimension);
          NpotPatch.resizeSystemIcon(shrinkRatio);

          const oldBlt = Bitmap.prototype.blt;

          Bitmap.prototype.blt = function (
            source,
            sx,
            sy,
            sw,
            sh,
            dx,
            dy,
            dw,
            dh
          ) {
            if (source._url === iconSetUrl) {
              dw = dw || sw / shrinkRatio[0];
              dh = dh || sh / shrinkRatio[1];
            }
            oldBlt.call(this, source, sx, sy, sw, sh, dx, dy, dw, dh);
          };
        }
      }
      return isReady;
    };
  }

  static applyPatch(level) {
    if (this.isPatched) return;
    console.log(`Applying NPOT Patch with level: ${level}`);
    
    if (level === "none") {
      this.isPatched = true;
      return;
    }
    
    this.patchBitmap();
    this.patchSprite();
    
    if (level === "high") {
      this.patchPixiNpot();
      this.patchSystemSpriteSets();
    }
    
    this.isPatched = true;
  }

  static setTrigger(level) {
    if (typeof Main !== "undefined") {
      const oldOnScriptLoad = Main.prototype.onScriptLoad;
      Main.prototype.onScriptLoad = function () {
        if (this.loadCount + 1 === this.numScripts) {
          NpotPatch.applyPatch(level);
        }
        oldOnScriptLoad.call(this);
      };
    } else {
      NpotPatch.applyPatch(level);
    }
  }
}

NpotPatch.isPatched = false;
NpotPatch.npotInfoCache = {};
NpotPatch.checkedParallaxRatio = {};
NpotPatch.dummyAnchor = document.createElement("a");

// Function to be called from Android
window.applyNpotPatch = function(level) {
  NpotPatch.setTrigger(level);
};