(function () {
  let pluginsName = PluginManager._scripts;

  // KMS Minimap performance fix
  if (pluginsName.includes("KMS_Minimap")) {
    try {
      let oldChange = SceneManager.changeScene;
      SceneManager.changeScene = function () {
        oldChange.call(this);
        try {
          SceneManager._scene.children[0]._minimap.children
            .filter((_) => _.updateBlink)
            .forEach((_) => (_.updateBlink = function () {}));
        } catch (e) {}
      };
    } catch (e) {
      console.error(e);
    }
  }

  // DKTools plugin fix
  if (pluginsName.includes("DKTools")) {
    try {
      DKTools.IO.isLocalMode = () => true;
    } catch (e) {
      console.error(e);
    }
  }

  // FTKR_GDM_WindowEditor plugin fix
  if (pluginsName.includes("FTKR_GDM_WindowEditor")) {
    try {
      Window_Base.prototype.spacing = function () {
        return 12;
      };
    } catch (e) {
      console.error(e);
    }
  }

  // MPP_MessageEx plugin fix
  if (pluginsName.includes("MPP_MessageEX")) {
    try {
      MPP.Window_MessageName.prototype.resetTextColor = function () {
        this.changeTextColor(this.textColor(this._colorIndex || 0));
      };
    } catch (e) {
      console.error(e);
    }
  }

  // MRCY_FacesetSystem_Temp plugin fix
  if (pluginsName.includes("MRCY_FacesetSystem_Temp")) {
    try {
      const rdCache = {};
      window.getDirsIn = function (baseDir) {
        log("[UTILITY] getDirsIn (" + baseDir + ")", DebugLevel.TRACE);
        baseDir = join(Core._gamePath, baseDir);
        let results = [];
        if (!rdCache[baseDir]) rdCache[baseDir] = readdirSync(baseDir, true);
        for (let file of rdCache[baseDir]) {
          if (file.isDirectory()) {
            results.push(file.name);
          }
        }

        log(
          "[getDirsIn] Found " + getArrayString(results) + " in " + baseDir,
          DebugLevel.FINER
        );
        return results;
      };

      window.getFilesIn = function (baseDir) {
        log("[UTILITY] getFilesIn (" + baseDir + ")", DebugLevel.TRACE);
        baseDir = join(Core._gamePath, baseDir);
        let results = [];
        if (!rdCache[baseDir]) rdCache[baseDir] = readdirSync(baseDir, true);
        for (let file of rdCache[baseDir]) {
          if (
            !file.isDirectory() &&
            (file.name.endsWith(".png") || file.name.endsWith(".rpgmvp"))
          ) {
            results.push(file.name);
          }
        }

        log(
          "[getFilesIn] Found " + getArrayString(results) + " in " + baseDir,
          DebugLevel.FINER
        );
        return results;
      };

      window.fileExists = function (fileName) {
        return true;
      };
    } catch (e) {
      console.error(e);
    }
  }

  if (pluginsName.includes("BattleMotion")) {
    Sprite_Battler.prototype.oneMotionFps = function (mainSprite) {
      var bitmap = mainSprite.bitmap;
      var cw = (bitmap.originalHeight || bitmap.height) / 6; //タテヨコサイズは固定するのでこれでいい
      var length = Object.keys(Sprite_Battler.MOTIONS).length;
      var oneMotion = (bitmap.originalWidth || bitmap.width) / (length / 6); //１ﾓｰｼｮﾝの大きさ
      var fpsMotion = oneMotion / cw; //１ﾓｰｼｮﾝの最大ﾌﾚｰﾑ数

      return fpsMotion;
    };

    Sprite_Actor.prototype.updateFrame = function () {
      if (this._mainSprite) {
        var bitmap = this._mainSprite.bitmap;
        if (bitmap.width <= 0) return;

        let [widthRatio, heightRatio] = [1, 1];
        if (bitmap.originalWidth && bitmap.originalHeight) {
          [widthRatio, heightRatio] = NpotPatch.getShrinkRatio([
            bitmap.originalWidth,
            bitmap.originalHeight,
          ]);
        }

        if (this._weaponSprite) {
          this._weaponSprite.visible = Sprite_Battler.drawWeapon;
        }

        var flg = this.getRemake();
        if (flg === true && bitmap.width !== 0) {
          this.addMotionData(this._mainSprite);
          this.setMotionFps(this._mainSprite);
          this.setMotionArray();
          this.setRemake(false);
        }

        var ch = this.cs(this._mainSprite) * heightRatio;
        var cw = this.cs(this._mainSprite) * widthRatio;
        var cx = this.cx();
        var cy = this.cy();

        this._mainSprite.setFrame(cx * cw, cy * ch, cw, ch);
      }
    };

    Sprite_Enemy.prototype.updateSVFrame = function () {
      Sprite_Battler.prototype.updateFrame.call(this);
      var bitmap = this._mainSprite.bitmap;
      if (!bitmap) return;
      if (bitmap.width <= 0) return;

      let [widthRatio, heightRatio] = [1, 1];
      if (bitmap.originalWidth && bitmap.originalHeight) {
        [widthRatio, heightRatio] = NpotPatch.getShrinkRatio([
          bitmap.originalWidth,
          bitmap.originalHeight,
        ]);
      }

      this._effectTarget = this._mainSprite;

      var flg = this.getRemake();
      if (flg === true && bitmap.width !== 0) {
        this.addMotionData(this._mainSprite);
        this.setMotionFps(this._mainSprite);
        this.setMotionArray();
        this.setRemake(false);
      }

      var ch = this.cs(this._mainSprite) * heightRatio;
      var cw = this.cs(this._mainSprite) * widthRatio;
      var cx = this.cx();
      var cy = this.cy();
      var cdh = 0;
      if (this._effectType === "bossCollapse") {
        cdh = ch - this._effectDuration * heightRatio;
      }
      // this.setFrame(cx * cw, cy * ch, cw, ch);
      this._mainSprite.setFrame(cx * cw, cy * ch, cw, ch - cdh);
      this.adjustMainBitmapSettings(bitmap);
      this.adjustSVShadowSettings();
    };
  }

  if (pluginsName.includes("PictureAnimation")) {
    try {
      const oldShowPicture = Game_Screen.prototype.showPicture;
      Game_Screen.prototype.showPicture = function (
        pictureId,
        name,
        origin,
        x,
        y,
        scaleX,
        scaleY,
        opacity,
        blendMode
      ) {
        oldShowPicture.apply(this, arguments);
        var realPictureId = this.realPictureId(pictureId);
        if (this._paCellNumber > 1) {
          this._pictures[realPictureId].setAnimationFrameInit(
            this._paCellNumber,
            this._paFrameNumber,
            this._paDirection,
            this._paFadeDuration,
            this._paSounds
          );
          this.clearPicturesAnimation();
        }
      };

      Sprite_Picture.prototype.updateAnimationFrame = function (
        sprite,
        cellCount
      ) {
        const gamePicture = this.picture();
        if (sprite.bitmap.originalWidth && sprite.bitmap.originalHeight) {
          const { originalWidth, originalHeight } = sprite.bitmap;

          if (!gamePicture._originalScaleX)
            gamePicture._originalScaleX = gamePicture._targetScaleX;
          if (!gamePicture._originalScaleY)
            gamePicture._originalScaleY = gamePicture._targetScaleY;
          gamePicture._scaleX = gamePicture._targetScaleX =
            (gamePicture._originalScaleX * originalWidth) / sprite.bitmap.width;
          gamePicture._scaleY = gamePicture._targetScaleY =
            (gamePicture._originalScaleY * originalHeight) /
            sprite.bitmap.height;
          if(sprite.updateScale) sprite.updateScale();
          //          $gameScreen.erasePicture(sprite._pictureId);
          //          $gameScreen.showPicture(
          //            sprite._pictureId,
          //            gamePicture._name,
          //            gamePicture._origin,
          //            gamePicture._x,
          //            gamePicture._y,
          //            (100 * originalWidth) / sprite.bitmap.width,
          //            (100 * originalHeight) / sprite.bitmap.height,
          //            gamePicture._opacity,
          //            gamePicture._blendMode
          //          );
        }
        switch (this.picture().direction()) {
          case "連番":
          case "N":
            sprite.bitmap = this._bitmaps[cellCount];
            sprite.setFrame(0, 0, sprite.bitmap.width, sprite.bitmap.height);
            break;
          case "縦":
          case "V":
            var height =
              (sprite.bitmap.originalHeight || sprite.bitmap.height) /
              this.picture().cellNumber();
            var y = cellCount * height;
            sprite.setFrame(0, y, sprite.bitmap.width, height);
            break;
          case "横":
          case "H":
            var width =
              (sprite.bitmap.originalWidth || sprite.bitmap.width) /
              this.picture().cellNumber();
            var x = cellCount * width;
            sprite.setFrame(x, 0, width, sprite.bitmap.height);
            break;
          default:
            sprite.setFrame(0, 0, this.bitmap.width, this.bitmap.height);
        }
      };
    } catch (e) {
      console.error(e);
    }
  }

  if (pluginsName.includes("CustomizeErrorScreen")) {
    Graphics._updateErrorPrinter = function () {
      const width = 640 * this._realScale;
      const height = 100 * this._realScale;
      this._errorPrinter.style.width = width + "px";
      this._errorPrinter.style.height = height + "px";
    };
  }

  if (pluginsName.includes("EventEffects")) {
    Sprite_Character.prototype.updateTone = function () {
      const tone = this._character.tone();
      if (tone instanceof Array) {
        if (!this._colorTone.equals(tone)) {
          this._colorTone = tone.slice();
        }
      }
    };
  }
})();
